// Pic.cpp

#include "Pic.h"

// ############
// Constructors
// ############
Pic::Pic ()
{
  // Set default values of calculation parameters
  N    = 20000;
  J    = 1000;
  L    = 100.;
  vb   = 3.;
  dt   = 0.1;
  tmax = 50.;
}

// ##########
// Destructor
// ##########
Pic::~Pic ()
{
}

// ###############
// Status function
// ###############
void Pic::Status ()
{ 
  printf ("\nN  = %6d       J    = %5d\n",
	  N, J);
  printf ("L  = %11.4e  vb   = %11.4e\n",
	  L, vb);
  printf ("dt = %11.4e  tmax = %11.4e\n\n",
	  dt, tmax);
}

// #######################################
// Functions to set calculation parameters
// #######################################
int Pic::SetN (int _N)
{
  if (_N < 1)
    {
      printf ("Pic::SetN: Error - invalid data value: N must be positive\n");
      return 1;
    }
  N = _N;
  return 0;
}
int Pic::SetJ (int _J)
{
  if (_J <= 1 || _J%2 != 0)
    {
      printf ("Pic::SetJ: Error - invalid data value: J must be even and positive\n");
      return 1;
    }
  J = _J;
  return 0;
}
int Pic::SetL (double _L)
{
  if (_L <= 0.)
    {
      printf ("Pic::SetL: Error - invalid data value: L must be positive\n");
      return 1;
    }
  L = _L;
  return 0;
}
int Pic::Setvb (double _vb)
{
  if (_vb <= 0.)
    {
      printf ("Pic::Setvb: Error - invalid data value: vb must be positive\n");
      return 1;
    }
  vb = _vb;
  return 0;
}
int Pic::Setdt (double _dt)
{
  if (_dt <= 0.)
    {
      printf ("Pic::Setdt: Error - invalid data value: dt must be positive\n");
      return 1;
    }
  dt = _dt;
  return 0;
}
int Pic::Settmax (double _tmax)
{
  if (_tmax <= 0.)
    {
      printf ("Pic::Settmax: Error - invalid data value: tmax must be positive\n");
      return 1;
    }
  tmax = _tmax;
  return 0;
}

// ######################################################################
// Function to evolve electron equations of motion from t = 0 to t = tmax
// ######################################################################
void Pic::Solve ()
{
  // Allocate solution vector
  double* Y = new double [2*N];
  
  // Seed random number generator
  int seed = time (NULL);
  srand (seed);
  
  // Initialize calculation
  FILE*  file = OpenFile ((char*) "Energy/energy.out");
  int    step = 0;
  double t    = 0.;
  for (int i = 0; i < N; i++)
    {
      Y[i]   = L * double (rand ()) /double (RAND_MAX);
      Y[N+i] = DoubleMaxwellian ();
    }
  WriteOutput (step, Y);

  // Evolve equations of motion
  double Kinetic, Total;
  do
    {
      // Take time step
      Evolve (t, Y);

      // Ensure that all electron coordinates lie in range 0 to L
      for (int i = 0; i < N; i++)
	{
	  if (Y[i] < 0.)
	    Y[i] += L;
	  if (Y[i] > L)
	    Y[i] -= L;
	}

      // Output data
      step++;
      CalcEnergy (Y, Kinetic, Total);
      WriteOutput (step, Y);
      fprintf (file, "%e %e %e\n", t, Kinetic, Total);
      printf ("Step = %3d  t = %11.4e\n", step, t);
    }
  while (t < tmax);
  fclose (file);
  printf ("\n");

  // Store final electron phase-space coordinates
  file = OpenFile ((char*) "phase-final.out");
  for (int i = 0; i < N; i++)
    fprintf (file, "%e %e\n", Y[i], Y[N+i]);
  fclose (file);
  
  // Clean up
  delete[] Y; 
}

// #########################################################
// Function to evolve electron equations of motion backwards
// #########################################################
void Pic::SolveBackwards ()
{
  // Allocate solution vector
  double* Y = new double [2*N];
  
  // Seed random number generator
  int seed = time (NULL);
  srand (seed);
  
  // Initialize calculation
  int    step = 0;
  double t    = 0.;
  FILE*  file = OpenFiler ((char*) "phase-final.out");
  for (int i = 0; i < N; i++)
    {
      double r, v;
      fscanf (file, "%lf %lf", &r, &v);
      Y[i]   = r;
      Y[N+i] = -v;
    }
  WriteOutput (step, Y);
  fclose (file);

  // Evolve equations of motion
  do
    {
      // Take time step
      Evolve (t, Y);

      // Ensure that all electron coordinates lie in range 0 to L
      for (int i = 0; i < N; i++)
	{
	  if (Y[i] < 0.)
	    Y[i] += L;
	  if (Y[i] > L)
	    Y[i] -= L;
	}

      // Output data
      step++;
      WriteOutput (step, Y);
      printf ("Step = %3d  t = %11.4e\n", step, t);
    }
  while (t < tmax);
  printf ("\n");
  
  // Clean up
  delete[] Y; 
}

// ######################################################
// Function to calculate total system energy per electron
// ######################################################
void Pic::CalcEnergy (double* Y, double& Kinetic, double& Total)
{
  // Allocate arrays
  double* r   = new double [N];
  double* v   = new double [N];
  double* ne  = new double [J];
  double* phi = new double [J];
  double* E   = new double [J];

  // Unload electron coordinates and velocities from solution vector
  UnLoadSolutionVector (Y, r, v);

  // Calculate fields
  CalcFields (r, ne, phi, E);

  // Calculate energy
  Kinetic = 0., Total = 0.;
  for (int i = 0; i < N; i++)
    {
      double phii;
      double dx = L /double (J);
      int    j  = int (r[i] /dx);
      double y  = r[i] /dx - double (j);

      if (j+1 == J)
	phii = phi[j] * (1. - y) + phi[0]   * y;
      else
	phii = phi[j] * (1. - y) + phi[j+1] * y;

      Kinetic += 0.5 * v[i]*v[i];
      Total   += 0.5 * v[i]*v[i] - 0.5 * phii;
    }
  Kinetic /= double (N);
  Total   /= double (N);

  // Clean up
  delete[] r; delete[] v; delete[] ne; delete[] phi; delete[] E;
}

// ##########################################
// Function to write output for nth time step
// ##########################################
void Pic::WriteOutput (int step, double *Y)
{
  // Allocate arrays
  double* r   = new double [N];
  double* v   = new double [N];
  double* ne  = new double [J];
  double* phi = new double [J];
  double* E   = new double [J];

  // Unload electron coordinates and velocities from solution vector
  UnLoadSolutionVector (Y, r, v);

  // Calculate fields
  CalcFields (r, ne, phi, E);
  
  // Construct filename for electron phase-space coordinates
  char* out = new char [100];
  sprintf (out, "Phase/phase-%03d.out", step);

  // Write electron phase-space coordinates
  FILE* file = OpenFile (out);
  for (int i = 0; i < N; i++)
    fprintf (file, "%e %e\n", r[i], v[i]);
  fclose (file);

  // Construct filename for electric field
  sprintf (out, "Electric/electric-%03d.out", step);

  // Write electric field
  file = OpenFile (out);
  double dx = L /double (J), x;
  for (int j = 0; j < J; j++)
    {
      x = double (j) * dx;
      fprintf (file, "%e %e\n", x, E[j]);
    }
  x = L;
  fprintf (file, "%e %e\n", x, E[0]);
  fclose (file);

  // Construct filename for electron number density
  sprintf (out, "Density/density-%03d.out", step);

  // Write electron number density
  file = OpenFile (out);
  double n0 = double (N) /L;
  for (int j = 0; j < J; j++)
    {
      x = double (j) * dx;
      fprintf (file, "%e %e\n", x, ne[j]/n0);
    }
  x = L;
  fprintf (file, "%e %e\n", x, ne[0]/n0);
  fclose (file);

  // Clean up
  delete[] out; delete[] r; delete[] v; delete[] ne; delete[] phi; delete[] E;
}

// ###############################################################
// Function to return random variable distributed according to
// function that is sum of two Maxwellians of thermal widths unity
// and mean values +vb and -vb.
// Uses rejection method.
// ##############################################################
double Pic::DoubleMaxwellian ()
{
  // Generate random velocity value in range -5*vb to +5*vb
  double vmin = - 5. * vb;
  double vmax = + 5. * vb;
  double rnd  = double (rand ()) /double (RAND_MAX);
  double v    = vmin + (vmax - vmin) * rnd;

  // Accept or reject value
  double fmax = 0.5 * (1. + exp (- 2. * vb*vb));
  double f    = 0.5 * exp (- (v - vb) * (v - vb) /2.);
  f          += 0.5 * exp (- (v + vb) * (v + vb) /2.);
  double g    = fmax * double (rand ()) /double (RAND_MAX);
  if (g > f)
    return DoubleMaxwellian ();
  else
    return v;
}

// ##############################################################################
// Function to integrate electron equations of motion from time t to time t + dt.
// Uses fourth-order Runge-Kutta algorithm.
// ##############################################################################
void Pic::Evolve (double& t, double* Y)
{
  // Allocate local arrays
  double* k1   = new double [2*N];
  double* k2   = new double [2*N];
  double* k3   = new double [2*N];
  double* k4   = new double [2*N];
  double* F    = new double [2*N];
  double* dYdt = new double [2*N];
 
  // Zeroth intermediate step 
  CalcRhs (Y, dYdt);
  for (int i = 0; i < 2*N; i++)
    {
      k1[i] = dt * dYdt[i];
      F[i]  = Y[i] + k1[i] /2.;
    }

  // First intermediate step 
  CalcRhs (F, dYdt);
  for (int i = 0; i < 2*N; i++)
    {
      k2[i] = dt * dYdt[i];
      F[i]  = Y[i] + k2[i] /2.;
    }

  // Second intermediate step
  CalcRhs (F, dYdt);
  for (int i = 0; i < 2*N; i++)
    {
      k3[i] = dt * dYdt[i];
      F[i]  = Y[i] + k3[i];
    }

  // Third intermediate step 
  CalcRhs (F, dYdt);
  for (int i = 0; i < 2*N; i++)
    k4[i] = dt * dYdt[i];

  // Actual step 
  for (int i = 0; i < 2*N; i++)
    Y[i] += k1[i] /6. + k2[i] /3. + k3[i] /3. + k4[i] /6.;
  t += dt;

  // Clean up
  delete[] k1; delete[] k2; delete[] k3; delete[] k4; delete[] F; delete[] dYdt;
}

// #####################################################################
// Function to calculate electron number density, electric potental, and
// electric field from solution vector. Does not modify solution vector
// #####################################################################
void Pic::CalcFields (double* r, double* ne, double* phi, double* E)
{
  // Memory allocation
  double* rho = new double [J];
  
  // Calculate electron number density
  Calcne (r, ne);

  // Calculate right-hand side of Poisson's equation
  double n0 = double (N) /L;
  for (int j = 0; j < J; j++)
    rho[j] = ne[j] /n0 - 1.;

  // Solve Poisson's equation to obtain electric potential
  Poisson_1D (phi, rho);

  // Calculate electric field from electric potential
  CalcE (phi, E);

  // Clean up
  delete[] rho;
}

// ######################################################################
// Function to calculate right-hand sides of electron equations of motion
//
// Y[0:N-1]  = r_i
// Y[N:2N-1] = v_i
//
// ######################################################################
void Pic::CalcRhs (double* Y, double* dYdt)
{
  // Memory allocation
  double* r    = new double [N];
  double* v    = new double [N];
  double* rdot = new double [N];
  double* vdot = new double [N];
  double* ne   = new double [J];
  double* phi  = new double [J];
  double* E    = new double [J];

  // Unload electron coordinates and velocities from solution vector
  UnLoadSolutionVector (Y, r, v);

  // Ensure that all electron coordinates lie in range 0 to L
  for (int i = 0; i < N; i++)
    {
      if (r[i] < 0.)
	r[i] += L;
      if (r[i] > L)
	r[i] -= L;
    }
  
  // Calculate electron number density, electric potential, and electric field
  CalcFields (r, ne, phi, E);
  
  // Evaluate right-hand sides of electron equations of motion
  for (int i = 0; i < N; i++)
    {
      // Calculate E(r_i)
      double Ei;
      double dx = L /double (J);
      int    j  = int (r[i] /dx);
      double y  = r[i] /dx - double (j);

      if (j+1 == J)
	Ei = E[j] * (1. - y) + E[0]   * y;
      else
	Ei = E[j] * (1. - y) + E[j+1] * y;

      // Set right-hand sides
      rdot[i] = v[i];
      vdot[i] = - Ei;
    }

  // Load right-hand sides of electron equations of motion into dYdt
  LoadSolutionVector (rdot, vdot, dYdt);
  
  // Clean up
  delete[] r;  delete[] v;   delete[] rdot; delete[] vdot;
  delete[] ne; delete[] phi; delete[] E;
}

// #########################################################################
// Function to load electron coordinates and velocities into solution vector
// #########################################################################
void Pic::LoadSolutionVector (double* r, double* v, double* Y)
{
  for (int i = 0; i < N; i++)
    {
      Y[i]   = r[i];
      Y[N+i] = v[i];
    }
}

// ###########################################################################
// Function to unload electron coordinates and velocities from solution vector
// ###########################################################################
void Pic::UnLoadSolutionVector (double* Y, double* r, double* v)
{
  for (int i = 0; i < N; i++)
    {
      r[i] = Y[i];
      v[i] = Y[N+i];
    }
}

// #######################################################################
// Function to calculate electron number density from electron coordinates
// #######################################################################
void Pic::Calcne (double* r, double* ne)
{
  // Initialize calculation
  double dx = L /double (J);
  
  for (int j = 0; j < J; j++)
    ne[j] = 0.;

  // Calculate number density via linear particle-in-cell algorithm
  for (int i = 0; i < N; i++)
    {
      int    j = int (r[i] /dx);
      double y = r[i] /dx - double (j);
      
      ne[j] += (1. - y) /dx;
      if (j+1 == J)
	ne[0]   += y /dx;
      else
	ne[j+1] += y /dx;
    }
}

// ############################################################
// Function to calculate electric field from electric potential
// ############################################################
void Pic::CalcE (double* phi, double* E)
{
   double dx = L /double (J);

   for (int j = 1; j < J-1; j++)
     E[j] = (phi[j-1] - phi[j+1]) /2./dx;

   E[0]   =  (phi[J-1] - phi[1]) /2./dx;
   E[J-1] =  (phi[J-2] - phi[0]) /2./dx;
 }

// ########################################################################
// Function to solve Poisson's equation in one dimension via Fourier method
//
// Solves
//       d^2 u /d x^2 = v for 0 <= x <= L.
// Periodic boundary conditions:
//       u(x+L) = u(x), v(x+L) = v(x).
// Arrays u and v assumed to be of length J.
// Grid points at
//       x_j = j dx for j = 0,J-1
// where
//       dx = L / J.
//
// ########################################################################
void Pic::Poisson_1D (double* u, double* v)
{
  // Memory allocation
  double* Vr = new double [J];
  double* Vi = new double [J];
  double* Ur = new double [J];
  double* Ui = new double [J];

  // Fourier transform v
  Fourier_Forward (v, Vr, Vi);

  // Calculate Fourier transform of u
  double kappa = 2.*M_PI /L;
  Ur[0] = 0.; Ui[0] = 0.;
  for (int j = 1; j <= J/2; j++)
    {
      Ur[j] = - Vr[j] /double (j*j) /kappa/kappa;
      if (j == J/2)
	Ui[j] = 0.;
      else
	Ui[j] = - Vi[j] /double (j*j) /kappa/kappa;
    }
  for (int j = J/2+1; j < J; j++)
    {
      Ur[j] =   Ur[J-j];
      Ui[j] = - Ui[J-j];
    }

  // Inverse Fourier transform to obtain u
  Fourier_Backward (Ur, Ui, u);

  // Clean up
  delete[] Vr; delete[] Vi; delete[] Ur; delete[] Ui;
}

// ################################################################################## 
// Function to calculate Fourier transform of real array f in complex array (Fr, Fi).
// Uses fftw.
// ##################################################################################
void Pic::Fourier_Forward (double* f, double* Fr, double* Fi)
{
  fftw_complex ff[J], FF[J];

  // Load data
  for (int j = 0; j < J; j++)
    {
      c_re (ff[j]) = f[j];
      c_im (ff[j]) = 0.;
    }

  // Call fftw routine
  fftw_plan p = fftw_create_plan (J, FFTW_FORWARD, FFTW_ESTIMATE);
  fftw_one (p, ff, FF);
  fftw_destroy_plan (p);

  // Unload data
  for (int j = 0; j < J; j++)
    {
      Fr[j] = c_re (FF[j]) /double (J);
      Fi[j] = c_im (FF[j]) /double (J);
    }
}

// ########################################################################################## 
// Function to calculate inverse Fourier transform of complex array (Fr, Fi) in real array f.
// Uses fftw.
// ##########################################################################################
void Pic::Fourier_Backward (double* Fr, double* Fi, double* f)
{
  fftw_complex ff[J], FF[J];

  // Load data
  for (int j = 0; j < J; j++)
    {
      c_re (FF[j]) = Fr[j];
      c_im (FF[j]) = Fi[j];
    }

  // Call fftw routine
  fftw_plan p = fftw_create_plan (J, FFTW_BACKWARD, FFTW_ESTIMATE);
  fftw_one (p, FF, ff);
  fftw_destroy_plan (p);

  // Unload data
  for (int j = 0; j < J; j++)
    f[j] = c_re (ff[j]);
 }

// #####################################
// Function to open new file for writing
// #####################################
FILE* Pic::OpenFile (char* filename)
{
  FILE* file = fopen (filename, "w");
  if (file == NULL) 
    {
      printf ("Pic::OpenFile: Error opening data-file: %s\n", filename);
      exit (1);
    }
  return file;
}

// ##########################################
// Function to open existing file for reading
// ##########################################
FILE* Pic::OpenFiler (char* filename)
{
  FILE* file = fopen (filename, "r");
  if (file == NULL) 
    {
      printf ("Pic::OpenFiler: Error opening data-file: %s\n", filename);
      exit (1);
    }
  return file;
}
 

